import asyncio
import json
import logging
import pickle
import uuid
from functools import partial
from typing import Any, Callable, Dict, List, Union

from aio_pika import DeliveryMode, Message
from aio_pika.abc import (
    AbstractIncomingMessage,
    AbstractRobustChannel,
    AbstractRobustExchange,
    AbstractRobustQueue,
    ConsumerTag,
)

from .constant import CANCEL_QUEUE_NAME
from .enums import ContentType
from .serializers import BaseSerializer

log = logging.getLogger(__name__)


class Client:
    tmp_queue: AbstractRobustQueue = None
    tmp_consumer: ConsumerTag = None

    def __init__(
        self, channel: AbstractRobustChannel, exchange: AbstractRobustExchange
    ) -> None:
        self.loop = asyncio.get_event_loop()
        self.channel = channel
        self.exchange = exchange
        self.futures: Dict[str, asyncio.Future] = {}
        self.serializers: List[BaseSerializer] = []

    @classmethod
    async def create(
        cls, channel: AbstractRobustChannel, exchange: AbstractRobustExchange
    ):
        instance = cls(channel, exchange)
        tmp_queue = await channel.declare_queue(exclusive=True, auto_delete=True)
        await tmp_queue.bind(exchange)
        tmp_consumer = await tmp_queue.consume(
            instance.on_message_received, no_ack=True
        )
        instance.tmp_queue = tmp_queue
        instance.tmp_consumer = tmp_consumer
        return instance

    def add_serializer(self, serializer: BaseSerializer):
        self.serializers.append(serializer)

    async def on_message_received(self, msg: AbstractIncomingMessage):
        cid = msg.correlation_id
        if cid is None:
            log.warn("Get messages without a correlation ID")
            return

        future = self.futures.pop(cid, None)
        if future is None:
            log.warning("Unknown message: %r", msg)
            return

        message = None
        try:
            success, result = pickle.loads(msg.body)
            if not success:
                future.set_exception(result)
                return

            for serializer in self.serializers:
                match = msg.content_type in serializer.content_type
                if match:
                    message = await serializer.deserialize(result)
                    break

            if message is None:
                raise TypeError(
                    f"Message are not supported. Serializer is not available for {msg.content_type!r}"
                )
        except Exception as e:
            log.error("Failed to deserialize message")
            future.set_exception(e)
            return

        future.set_result(message)

    def create_future(self):
        future = self.loop.create_future()
        cid = str(uuid.uuid4())
        self.futures[cid] = future
        future.add_done_callback(partial(self._remove_future, cid))
        return cid, future

    def _remove_future(self, cid: str, f: asyncio.Future):
        log.debug("Removing future: " + cid)
        self.futures.pop(cid, None)

    async def call(
        self,
        func: Union[Callable, str],
        *,
        args: List[Any] = [],
        kwargs: Dict[str, Any] = {},
        timeout: float = None,
        **msg_kwargs,
    ):
        if callable(func):
            func = func.__name__

        if not isinstance(func, str):
            raise ValueError(
                "func parameter must be a string. (this is the same as the function name)"
            )

        params = {"args": args, "kwargs": kwargs}
        body = json.dumps(params).encode()
        cid, future = self.create_future()
        msg_kwargs.setdefault("content_type", ContentType.TEXT)
        msg_kwargs["correlation_id"] = cid
        msg_kwargs["delivery_mode"] = DeliveryMode.NOT_PERSISTENT
        msg_kwargs["reply_to"] = self.tmp_queue.name
        message = Message(body, **msg_kwargs)
        await self.exchange.publish(message, func)
        log.info(f"Message has been sent to: {func!r}")
        log.debug("Waiting for the result...")
        try:
            result = await asyncio.wait_for(future, timeout)
            log.debug(f"Get results from {func!r}: {result}")
        except asyncio.TimeoutError:
            msg = Message(b"", correlation_id=cid)
            await self.exchange.publish(msg, routing_key=CANCEL_QUEUE_NAME)
            log.debug(f"Cancel task: {cid!r}")
            raise

        return result

    async def close(self):
        if self.tmp_queue and self.tmp_consumer:
            await self.tmp_queue.cancel(self.tmp_consumer)
            await self.tmp_queue.delete(if_unused=False, if_empty=False)



import json
from abc import ABC, abstractmethod
from typing import List, Union

from .enums import ContentType


class BaseSerializer(ABC):
    content_type: List[Union[ContentType, None]] = None

    @abstractmethod
    async def serialize(self, data: bytes):
        raise NotImplementedError

    @abstractmethod
    async def deserialize(self, data: bytes):
        raise NotImplementedError


class RawSerializer(BaseSerializer):
    content_type = [ContentType.TEXT, None]

    async def serialize(self, data: Union[str, bytes]):
        return data

    async def deserialize(self, data: bytes):
        return data


class JSONSerializer(BaseSerializer):
    content_type = [ContentType.JSON]

    async def serialize(self, data: Union[dict, list, int, float, str]) -> bytes:
        return json.dumps(data).encode()

    async def deserialize(self, data: bytes):
        return json.loads(data)



import asyncio
import json
import logging
import pickle
from functools import partial
from typing import Callable, Dict, List

from aio_pika import Message
from aio_pika.abc import (
    AbstractIncomingMessage,
    AbstractQueue,
    AbstractRobustChannel,
    AbstractRobustExchange,
    ConsumerTag,
)

from .constant import CANCEL_QUEUE_NAME
from .serializers import BaseSerializer

log = logging.getLogger(__name__)


class Server:
    cancel_queue: AbstractQueue = None
    cancel_queue_consumer: ConsumerTag = None

    def __init__(
        self,
        channel: AbstractRobustChannel,
        exchange: AbstractRobustExchange,
        *,
        prefix: str = "",
    ) -> None:
        self.loop = asyncio.get_event_loop()
        self.channel = channel
        self.exchange = exchange
        self.prefix = prefix
        self.functions: Dict[str, Callable] = {}
        self.consumers: Dict[str, ConsumerTag] = {}
        self.queues: Dict[str, AbstractQueue] = {}
        self.serializers: List[BaseSerializer] = []
        self.tasks: Dict[str, asyncio.Task] = {}

    @classmethod
    async def create(
        cls, channel: AbstractRobustChannel, exchange: AbstractRobustExchange
    ):
        instance = cls(channel, exchange)
        instance.cancel_queue = await channel.declare_queue(
            CANCEL_QUEUE_NAME, durable=True
        )
        await instance.cancel_queue.bind(exchange)
        instance.cancel_queue_consumer = await instance.cancel_queue.consume(
            instance.task_cancel_handler, no_ack=True
        )
        return instance

    async def task_cancel_handler(self, msg: AbstractIncomingMessage):
        cid = msg.correlation_id
        if not cid:
            log.warn("Get messages without a correlation ID")
            return

        task = self.tasks.get(cid)
        if task:
            if task.done():
                log.warn(f"Unable to cancel task: {cid!r} (Task completed)")
            else:
                task.cancel()
                log.info(f"Canceling the task succeeded: {cid!r}")
        else:
            log.error(f"Unable to cancel task: {cid!r} (Task not found)")

    def _remove_task(self, cid: str, t: asyncio.Task):
        self.tasks.pop(cid, None)
        log.debug(f"Task removed: {cid!r}")

    def add_serializer(self, serializer: BaseSerializer):
        self.serializers.append(serializer)

    async def add_route(self, routing_key: str, func: Callable, **queue_kwargs):
        if routing_key in self.functions:
            raise RuntimeError(f"function already registered: {func}")

        if not asyncio.iscoroutinefunction(func):
            raise RuntimeError(f"function should be coroutine: {func}")

        queue_kwargs["auto_delete"] = True
        queue_name = self.prefix + routing_key
        queue = await self.channel.declare_queue(queue_name, **queue_kwargs)
        await queue.bind(self.exchange, routing_key)
        self.queues[routing_key] = queue
        self.functions[routing_key] = func
        self.consumers[routing_key] = await queue.consume(
            partial(self.on_message_received, routing_key), no_ack=True
        )
        log.info(f"Added: {routing_key!r}")

    async def on_message_received(self, routing_key: str, msg: AbstractIncomingMessage):
        func = self.functions.get(routing_key)
        if not func:
            log.warn(f"function for route {routing_key!r} not found")
            return

        if msg.reply_to is None:
            log.warn("Cannot find the reply-to attribute on the message.")
            return

        log.debug("Parse parameters...")
        func_params = await self.parse_params(msg)
        if not isinstance(func_params, tuple):
            return

        args, kwargs = func_params

        async def _func_executor():
            log.debug(f"Call function: {func.__name__!r}")
            success = False
            try:
                result = await func(*args, **kwargs)
            except Exception as e:
                log.exception(f"Error occurred when calling {func.__name__!r} function")
                result = e
            else:
                message_data = None
                try:
                    for serializer in self.serializers:
                        if msg.content_type in serializer.content_type:
                            message_data = await serializer.serialize(result)
                            break

                    if message_data is None:
                        raise TypeError(
                            f"Result from {func!r} are not supported. Serializer is not available for {msg.content_type!r}"
                        )

                    success = True
                    result = message_data
                except Exception as e:
                    result = e

            payload = pickle.dumps((success, result), protocol=pickle.HIGHEST_PROTOCOL)
            message = Message(payload)
            for msg_attr, msg_attr_value in msg.info().items():
                setattr(message, msg_attr, msg_attr_value)

            await self.exchange.publish(message, routing_key=msg.reply_to)
            log.info(f"Result have been forwarded to: {msg.reply_to!r}")

        task = asyncio.create_task(_func_executor())
        task.add_done_callback(partial(self._remove_task, msg.correlation_id))
        self.tasks[msg.correlation_id] = task

    async def parse_params(self, msg: AbstractIncomingMessage):
        try:
            params: dict = json.loads(msg.body)
            if not isinstance(params, dict):
                log.error(
                    f"The function parameter should be of type dict, not {type(params)}."
                )
                return
        except json.JSONDecodeError:
            log.exception(f"Unable to parse function parameters: {msg.body}")
            return

        args_param = params.get("args", [])
        kwds_param = params.get("kwargs", {})
        return args_param, kwds_param

    async def close(self):
        for routing_key, consumer_tag in self.consumers.items():
            queue = self.queues[routing_key]
            await queue.cancel(consumer_tag)
            await queue.delete(if_unused=False, if_empty=False)

        self.consumers.clear()
        self.queues.clear()
        if self.cancel_queue and self.cancel_queue_consumer:
            await self.cancel_queue.cancel(self.cancel_queue_consumer)
            await self.cancel_queue.delete(if_unused=False, if_empty=False)

        self.cancel_queue = None
        self.cancel_queue_consumer = None
        for task in self.tasks.values():
            if not task.done():
                task.cancel()

        self.tasks.clear()
        log.debug("Cleaned!")
